//================================================================================================================
//
// Files which must be included before including this file:
//
// sky_render geometric_functions.inc
// functions.inc
//
//
// Constants which must be declared before including this file:
//
// bool calc_secondary_scatter
// float this_frame_wavelgth
// float earth_r
// float atm_top_altitude
// float camera_altitude
// float sun_theta
// float sun_steradian
// float factor_mie_anisotropic1
// float factor_mie_anisotropic2
// float ground_albedo
// int sky_light_primary_path_res
// int sky_light_secondary_spl_res
// int sky_light_secondary_path_res
// int sun_extinct_res
// int skylight_extinct_res
//
//
// Functions which must be declared before including this file:
//
// f_rayleigh_extinction_coefficient()
// f_mie_extinction_coefficient1()
// f_mie_extinction_coefficient2()
// f_absorb_extinction_coefficient()
//
//================================================================================================================

#declare GOLDEN_RATIO = (1+sqrt(5))/2;
#declare lambda_dither_res = project_wavelgth_resolution;






//================================================================================================================
// Calculate the transmittance between two points in atmosphere,
// input the altitudes of these two points and the distance between them.
//================================================================================================================

#declare f_optical_dpth_incre =
function(_incre_dist, _altitude){
_incre_dist
*(
f_rayleigh_extinction_coefficient(_altitude, this_frame_wavelgth) + f_mie_extinction_coefficient1(_altitude) + f_mie_extinction_coefficient2(_altitude) + f_absorb_extinction_coefficient(_altitude)
)
}

#declare f_sky_transmittance_internal =
function(_r1, _r2, _incre_dist, _para_d, _res){
exp(
-sum(
l, 0, _res-1,
f_optical_dpth_incre(
_incre_dist,
sqrt(_r1*_r1 + 2*(1/2+l)*_incre_dist*_para_d + pow((1/2+l)*_incre_dist,2)) - earth_r
)
)
)
}

#declare f_sky_transmittance = //interface
function(_altitude1, _altitude2, _distance, _res){ // prerequisite: _distance > abs(_altitude2-_altitude1)  and  0 < _altitudes < atm_top_altitude
f_sky_transmittance_internal(
earth_r + _altitude1,
earth_r + _altitude2,
abs(_distance/_res),
(pow(earth_r+_altitude2,2) - pow(earth_r+_altitude1,2) - pow(_distance,2))/_distance/2,
_res
)
}



//================================================================================================================
// Calculate the (scattering)extinction coefficient and the phase function product for a given situation,
// input the altitude of the occurrance, and the cosine of the scattering deflection angle. Output unit 1/km/sr
//================================================================================================================

#declare f_scatter_factor =
function(_altitude, _between_cos){
f_rayleigh_extinction_coefficient(_altitude, this_frame_wavelgth)
*(1 + pow(_between_cos,2))*3/(16*pi)
+f_mie_extinction_coefficient1(_altitude)
*(1-pow(factor_mie_anisotropic1,2))*(1+pow(_between_cos,2))/(2+pow(factor_mie_anisotropic1,2))/pow(1+pow(factor_mie_anisotropic1,2)-2*factor_mie_anisotropic1*_between_cos,3/2)*3/(8*pi)
+f_mie_extinction_coefficient2(_altitude)
*(1-pow(factor_mie_anisotropic2,2))*(1+pow(_between_cos,2))/(2+pow(factor_mie_anisotropic2,2))/pow(1+pow(factor_mie_anisotropic2,2)-2*factor_mie_anisotropic2*_between_cos,3/2)*3/(8*pi)
}



//================================================================================================================
// Calculate the sun light sp irradiance in outter space, output unit W/m^2/m
//================================================================================================================

#declare f_sun_sp_irradiance =
function(_sp_radiance){
_sp_radiance*sun_steradian
}



//================================================================================================================
// Calculate the sun light sp irradiance in the atmosphere,
// input the absolute coordinate r, theta, phi of a given point. Output unit W/m^2/m
//================================================================================================================

#declare f_sun_light_field_internal =
function(_point_r, _point_theta, _point_phi, _point_to_atm_top_thck){ //_point_phi
select(
_point_to_atm_top_thck,
0,
f_sun_sp_irradiance(f_sun_sp_radiance(this_frame_wavelgth))
*f_sky_transmittance(atm_top_altitude, f_point_altitude(_point_r, _point_theta), _point_to_atm_top_thck, sun_extinct_res)
)
}

#declare f_sun_light_field =
function(_point_r, _point_theta, _point_phi){
f_sun_light_field_internal(
_point_r, _point_theta, _point_phi,
f_atmos_thck_from_point(_point_r, _point_theta, f_camera_theta_to_geoc_theta(_point_r, _point_theta), _point_phi, sun_theta, 0, atm_top_altitude)
)
}




//================================================================================================================
// Calculate the ground reflect sp radiance. The function design is performance orientation so it input the
// coordinate r, x, and z instead of r, theta, and phi. And call the f_sun_light_field_internal() directly.
// Output unit is W/m^2/sr/m
//================================================================================================================

#declare f_ground_radiance =
function(_point_r, _point_x, _point_z){
f_sun_light_field_internal(
_point_r,
acos(_point_z/_point_r),
0,
f_secondary_calc_sun_atmos_thck(_point_r, acos(_point_z/_point_r), f_camera_theta_to_geoc_theta(_point_r, acos(_point_z/_point_r)), _point_x/sqrt(_point_r*_point_r - _point_z*_point_z))
)
*(sin(f_camera_theta_to_geoc_theta(_point_r, acos(_point_z/_point_r)))*sin(sun_theta)*_point_x/sqrt(_point_r*_point_r - _point_z*_point_z) + cos(f_camera_theta_to_geoc_theta(_point_r, acos(_point_z/_point_r)))*cos(sun_theta))
*ground_albedo/pi
}





//================================================================================================================
// Calculate one sample of secondary light contribute to the primary path.
// Input the absolute coordinate r, theta, and phi of the sampling point, absolute theta and absolute phi of the
// sampling ray direction, the altitude of the sampling point for acceleration.
// Output unit is W/m^2/sr/km/m
//================================================================================================================

#declare f_sky_secondary_piece_radiance =
function(_point_r, _point_x, _point_z, _last_node_altitude, _dist_to_last_node, _secondary_path_lgth, _sun_light_cos){
f_sun_light_field_internal(
_point_r,
acos(_point_z/_point_r),
0,
f_secondary_calc_sun_atmos_thck(
    _point_r,
    acos(_point_z/_point_r),
    f_camera_theta_to_geoc_theta(_point_r, acos(_point_z/_point_r)),
    _point_x/sqrt(_point_r*_point_r - _point_z*_point_z)
)
)
*f_scatter_factor(f_point_altitude(_point_r, acos(_point_z/_point_r)), _sun_light_cos)
*abs(_secondary_path_lgth)/sky_light_secondary_path_res
*f_sky_transmittance( f_point_altitude(_point_r, acos(_point_z/_point_r)), _last_node_altitude, _dist_to_last_node, skylight_extinct_res)
}

#declare f_sky_one_secondary_path_radiance =
function(_point_r, _point_theta, _point_phi, _view_theta, _view_phi, _base_altitude, _secondary_path_lgth, _sun_light_cos){
sum(
j2, 0, sky_light_secondary_path_res-1,
f_sky_secondary_piece_radiance(
f_secondary_calc_point_r(_point_r, _point_theta, _point_phi, (1/2+j2)*abs(_secondary_path_lgth)/sky_light_secondary_path_res, _view_theta, _view_phi),
f_secondary_calc_point_x(_point_r, _point_theta, _point_phi, (1/2+j2)*abs(_secondary_path_lgth)/sky_light_secondary_path_res, _view_theta, _view_phi),
f_secondary_calc_point_z(_point_r, _point_theta, (1/2+j2)*abs(_secondary_path_lgth)/sky_light_secondary_path_res, _view_theta),
_base_altitude,
(1/2+j2)*abs(_secondary_path_lgth)/sky_light_secondary_path_res,
_secondary_path_lgth,
_sun_light_cos
)
)
#if(ground_albedo > 0)
    +select(
    _secondary_path_lgth,
    f_ground_radiance(
    f_secondary_calc_point_r(_point_r, _point_theta, _point_phi, abs(_secondary_path_lgth), _view_theta, _view_phi),
    f_secondary_calc_point_x(_point_r, _point_theta, _point_phi, abs(_secondary_path_lgth), _view_theta, _view_phi),
    f_secondary_calc_point_z(_point_r, _point_theta, abs(_secondary_path_lgth), _view_theta)
    )
    *f_sky_transmittance(0, _base_altitude, abs(_secondary_path_lgth), skylight_extinct_res),
    0
    )
#end
}

#declare f_sky_recieve_from_one_secondary_path = // interface
function(_point_r, _point_theta, _point_phi, _view_theta, _view_phi, _altitude){
f_sky_one_secondary_path_radiance(
    _point_r,
    _point_theta,
    _point_phi,
    _view_theta,
    _view_phi,
    _altitude,
    f_atmos_thck_from_point(_point_r, _point_theta, f_camera_theta_to_geoc_theta(_point_r, _point_theta), _point_phi, _view_theta, _view_phi, atm_top_altitude),
    f_cos_between_direct_sun_light(_view_theta, cos(_view_phi))
)
*4*pi/sky_light_secondary_spl_res
*f_scatter_factor(_altitude, sin(_point_theta)*sin(_view_theta)*(cos(_point_phi)*cos(_view_phi) + sin(_point_phi)*sin(_view_phi)) + cos(_point_theta)*cos(_view_theta))
}




//================================================================================================================
// Calculate one segment sp radiance of primary path. Both sun light and secondary sky light contribution will be
// taken. The secondary scattering direction is sampling here.
// Input the absolute coordinate r, theta, and phi of the segment mid point, the altitude of the same point for
// accelaration, length of the primary path it belong, and a dither.
// Output unit is W/m^2/sr/m
//================================================================================================================

#declare f_sky_piece_radiance_camera_view = //f_sky_primary_piece_radiance_to_camera
function(_camera_view_r, _camera_view_theta, _camera_view_phi, _altitude, _primary_path_lgth, _dith){
(
//sun light
    f_sun_light_field(_camera_view_r, _camera_view_theta, _camera_view_phi)
    *f_scatter_factor(_altitude, f_cos_between_direct_sun_light(_camera_view_theta, cos(_camera_view_phi)))
//secondary sky light
    +sum(
    j1, 0, sky_light_secondary_spl_res-1,
    f_sky_recieve_from_one_secondary_path(
        _camera_view_r, _camera_view_theta, _camera_view_phi,
        (cos(pi*f_granite((sky_light_secondary_spl_res*_dith+j1)*1000/sky_light_primary_path_res/sky_light_secondary_spl_res/project_wavelgth_resolution,0,0))+1)*pi/2,
        (sky_light_secondary_spl_res*_dith+j1)*2*pi*(1-1/GOLDEN_RATIO),
        _altitude
    )
    )
)
*_primary_path_lgth/sky_light_primary_path_res
*f_sky_transmittance(_altitude, camera_altitude, _camera_view_r, skylight_extinct_res)
}




//================================================================================================================
// Calculate the sp radiance of a primary path.
// Input the absolute coordinate theta, and phi, it just the camera view theta and the camera view phi for our
// propose. Output unit is W/m^2/sr/m
//================================================================================================================

#if(calc_secondary_scatter)
    #declare f_sky_radiance_camera_view_internal =
    function(_camera_view_theta, _camera_view_phi, _atmos_thck){  // _atmos_thck n[ȶ?
    sum(
    i, 0, sky_light_primary_path_res - 1,
    f_sky_piece_radiance_camera_view(
    (1/2+i)*_atmos_thck/sky_light_primary_path_res, _camera_view_theta, _camera_view_phi,
    f_point_altitude((1/2+i)*_atmos_thck/sky_light_primary_path_res, _camera_view_theta),
    _atmos_thck,
    i + sky_light_primary_path_res*(this_frame_wavelgth-project_min_wavelgth)/project_wavelgth_step
    )
    )    
    }

#else
    #declare f_sky_radiance_camera_view_internal =
    function(_camera_view_theta, _camera_view_phi, _atmos_thck){
    sum(
    i, 0, sky_light_primary_path_res - 1,
    f_sun_light_field((1/2+i)*_atmos_thck/sky_light_primary_path_res, _camera_view_theta, _camera_view_phi)
    *f_scatter_factor(f_point_altitude((1/2+i)*_atmos_thck/sky_light_primary_path_res, _camera_view_theta), f_cos_between_direct_sun_light(_camera_view_theta, cos(_camera_view_phi)))
    *_atmos_thck/sky_light_primary_path_res
    *f_sky_transmittance(f_point_altitude((1/2+i)*_atmos_thck/sky_light_primary_path_res, _camera_view_theta), camera_altitude, (1/2+i)*_atmos_thck/sky_light_primary_path_res, skylight_extinct_res)
    )
    }

#end

#declare f_sky_radiance_camera_view =
function(_camera_view_theta, _camera_view_phi){
f_sky_radiance_camera_view_internal(_camera_view_theta, _camera_view_phi, abs(f_atmos_thck_from_point(0, 0, 0, 0, _camera_view_theta, 0, atm_top_altitude)))

}
